#include <Poisson1.h>
#include <readOrMakeGrid.h>
#include <ElmMatVec.h>
#include <FiniteElement.h>
#include <ErrorNorms.h>
#include <Mat_real.h>

// functor for the analytical solution:
class El1AnalSol : public FieldFunc
{
  Poisson1* data;      // access to simulator
public:
  El1AnalSol (Poisson1* data_) { data = data_; }
  virtual real valuePt (const Ptv(real)& x, real /*t*/ = DUMMY);
};

real El1AnalSol:: valuePt (const Ptv(real)& x, real /*t*/)
{
  const int nsd = x.size();
  real sum = 0;
  for (int j = 1; j <= nsd; j++)
    sum += data->A(j)*sin(j*M_PI*x(j)); // M_PI=3.14...see cmath
  return sum;
}
// Alternative: El1AnalSol can have an internal Ptv(real) A, which is
// initialized in the constructor


Poisson1:: Poisson1 () {}

void Poisson1:: adm (MenuSystem& menu)  // administer the menu
{
  SimCase::attach (menu); // enables later access to menu
  define (menu);          // define/build the menu
  menu.prompt();          // prompt user, read menu answers into memory
  scan ();                // read menu answers into class variables and init
}

void Poisson1:: define (MenuSystem& menu, int level)
{
  menu.addItem (level,          // menu level: level+1 is a submenu
                "gridfile",     // menu command/name
                "file or preprocessor command",  // help/description
                "mygrid.grid"); // default answer

  menu.addItem (level,          // menu level: level+1 is a submenu
                "A parameters", // menu command/name
                "A_1 A_2 ... in f expression", // help/description
                "3 2 1");       // default answer
  menu.addItem(level, "standard calc element matrix", "std_elmat", 
		      "standard calc element matrix", "true", "S/true/false/ON/OFF"); 

  menu.addItem(level, "dump matrix", "dump", 
		      "dump the whole matrix", "true", "S/true/false/ON/OFF"); 
  // submenus:
  LinEqAdmFE::defineStatic (menu, level+1);  // linear solver
  FEM::defineStatic (menu, level+1);         // num. integr. rule etc
  SaveSimRes::defineStatic (menu, level+1);  // storage of fields
}

void Poisson1:: scan ()
{
  MenuSystem& menu = SimCase::getMenuSystem();
  // load answers from the menu:
  String gridfile = menu.get ("gridfile");  // menu.get returns a string
  grid.rebind (new GridFE());  // make an empty grid
  readOrMakeGrid (*grid, gridfile);
  A.redim (grid->getNoSpaceDim());
  A.scan (menu.get ("A parameters"));
  database.rebind (new SaveSimRes ());
  database->scan (menu, grid->getNoSpaceDim());
  FEM::scan (menu);

  standard_calcelmat = menu.get("standard calc element matrix").getBool(); 
  s_o <<"standard_calcelmat "<<standard_calcelmat<<endl; 
  dump = menu.get("dump matrix").getBool(); 
  s_o <<"dump "<<dump<<endl; 

  // allocate data structures in the class:
  u.rebind (new FieldFE (*grid,"u"));              // allocate, field name "u"
  error.rebind (new FieldFE (*grid, "error"));     // another scalar field
  flux.rebind (new FieldsFE (*grid, "flux"));      // vector field

  dof.rebind (new DegFreeFE (*grid, 1));   // 1 for 1 unknown per node
  lineq.rebind (new LinEqAdmFE());         // make linear system and solvers
  lineq->scan (menu);                      // determine storage and solver type
  linsol.redim (dof->getTotalNoDof());
  lineq->attach (linsol);             
  uanal.rebind (new El1AnalSol (this));    // functor for analytical solution

}

void Poisson1:: fillEssBC ()
{
  dof->initEssBC ();                       // init for assignment below
  const int nno = grid->getNoNodes();      // no of nodes
  Ptv(real) x;                             // coord. of a nodal point
  for (int i = 1; i <= nno; i++) {
    // is node i subjected to any boundary indicator?
    if (grid->boNode (i)) {
      x = grid->getCoor (i);               // extract coord. of node i
      dof->fillEssBC (i, g(x));            // u=g at boundary nodes
    }
  }
}

void Poisson1:: integrands (ElmMatVec& elmat, const FiniteElement& fe)
{
  int i,j,q;
//  real f_value = f (fe); 
//  real k_value = k (fe);
  const int nbf = fe.getNoBasisFunc(); // no of nodes (or basis functions)
  const real detJxW = fe.detJxW();     // det J times numerical itg.-weight
  const int nsd = fe.getNoSpaceDim();

//  s_o <<"detJ "<<fe.detJ()<<endl; 

  real gradNi_gradNj;
  for (i = 1; i <= nbf; i++) {
    for (j = 1; j <= nbf; j++) {
      gradNi_gradNj = 0;
      for (q = 1; q <= nsd; q++)
        gradNi_gradNj += fe.dN(i,q) * fe.dN(j,q);

      elmat.A(i,j) += gradNi_gradNj*detJxW;
    }
//    elmat.b(i) += fe.N(i)*f_value*detJxW;
  }
}

void Poisson1:: solveProblem () // main routine of class Poisson1
{
  s_o << "\nNo of nodes: " << grid->getNoNodes() << ", no of elements: "
      << grid->getNoElms() << ", element: " << grid->getElmType(1) << "\n";

//  fillEssBC ();                // set essential boundary conditions

  makeSystem (*dof, *lineq);   // calculate linear system

  if ( dump ) {
    String S = "A_" + casename +".m"; 
    lineq->A().save(S.c_str(), "A"); 
  }

  /*
  linsol.fill (0.0);           // set all entries to 0 in start vector
  dof->insertEssBC (linsol);   // insert boundary values in start vector
                               // (improves linsol when g != 0)
  lineq->solve();              // solve linear system

  int niterations; bool c;  // for iterative solver statistics
  if (lineq->getStatistics(niterations,c))  // iterative solver?
    s_o << oform("  solver%sconverged in %3d iterations\n",
                 c ? " " : " not ",niterations);

  // the solution is now in linsol, it must be copied to the u field:
  dof->vec2field (linsol, *u);

  database->dump (*u);         // dump u for later visualization
  database->lineCurves (*u);   // dump u along lines for later plotting

  FEM::makeFlux  (*flux, *u);  // compute -k*grad(u) and smooth the result
  database->dump (*flux);      // dump vector field for later visualization
  */
}

void Poisson1:: resultReport ()
{
  /*
  // find error field and error norms:
  ErrorNorms::errorField (*uanal, *u, DUMMY, *error);
  database->dump (*error);
  ErrorNorms::Lnorm (*uanal,        // supplied function (see above)
		     *u,            // numerical solution
		     DUMMY,         // point of time
		     L1_error, L2_error, Linf_error, // error norms
		     GAUSS_POINTS); // point type for numerical integr.
  s_o << oform("  L1-error=%12.5e,  L2-error=%12.5e,  max-error=%12.5e\n",
               L1_error, L2_error, Linf_error);
  */
}

real Poisson1:: f (const FiniteElement& fe)
{
  Ptv(real) x = fe.getGlobalEvalPt();
  const int nsd = x.size();
  real sum = 0;
  for (int j = 1; j <= nsd; j++)
    sum += sqr(j*M_PI)*A(j)*sin(j*M_PI*x(j));
  return sum;
}

real Poisson1:: g (const Ptv(real)& x) { return uanal->valuePt (x); }
// The k function is used in integrands in defining the weak form *and* in
// FEM::makeFlux when computing the flux -k*grad(u). The FEM::makeFlux
// function requires that k here in the solver class has exactly the
// following signature (real return value and FiniteElement and real arguments)
real Poisson1:: k (const FiniteElement& /*fe*/, real /*t*/) { return  1; }


void Poisson1:: calcElmMatVec (int e, ElmMatVec& elmat, FiniteElement& fe){
  if ( standard_calcelmat) {
    FEM::calcElmMatVec(e,elmat,fe); 
  } else {
    if (fe.getNoSpaceDim() == 2 )  SyFicalcElmMatVec2D(e,elmat, fe); 
    else if (fe.getNoSpaceDim() == 3 )  SyFicalcElmMatVec3D(e,elmat, fe); 
  }
//  elmat.print(s_o); 
}


