#include <MatrixFactory_highorder.h>
#include <ElementMatrixCalculator.h>
#include <compute_conductivity.h>
#include <iostream>
#include <utility>
#include <map>
#include <vector>
#include <iterator>



using namespace dolfin;
using namespace std;

MatrixFactory:: MatrixFactory(Mesh* mesh_){
  mapfac = new MapMatrixFactory(mesh_); 
  // default order  
  order1 = 1; 
  order2 = 1; 
}


FastMatSparse* MatrixFactory:: computeMassMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeMassMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeMassMatrix(ScalarFunction& func, bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeMassMatrix(func,append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}



FastMatSparse* MatrixFactory:: computeStiffnessMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeStiffnessMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeStiffnessMatrix(TensorFunction& tf, bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeStiffnessMatrix(tf, append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeConvectionMatrix(VectorFunction& func,bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeConvectionMatrix(func,append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeVectorMassMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorMassMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeVectorStiffnessMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorStiffnessMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeVectorStiffnessMatrix(TensorFunction& tf, bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorStiffnessMatrix(tf, append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeVectorConvectionMatrix(double* v, int size,bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorConvectionMatrix(v,size,append_to_matrix); 
  idof = mapfac->idof; 
  return new FastMatSparse(*(mapfac->matrix));  
}

FastMatSparse* MatrixFactory:: computeDivMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->order2 = order2; 
  mapfac->computeDivMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  jdof = mapfac->jdof; 
  return new FastMatSparse(*(mapfac->matrix),true);  
}




MapMatrixFactory:: MapMatrixFactory(Mesh* mesh_) {
    mesh = mesh_;
    matrix = new map<pair<int,int>,double>();
    rhs = new map<int,double>(); 
    idof = new Dof_ptv(true,true); 
    jdof = new Dof_ptv(true,true); 
    order1 = 1; 
    order2 = 1; 
}

void dump_to_matlab(std::map<pair<int,int>,double> A, const char *varname, const char *s)
{
  ofstream outfile(s, ios::out);
  outfile << "%The matrix\n"; 
  std::map<std::pair<int,int>,double>::const_iterator Ait;
  double eps = 1.0e-10; 

  for(Ait=A.begin(); Ait!=A.end(); Ait++){
    outfile << varname << "(" << (*Ait).first.first+1  << ","
            << (*Ait).first.second+1  << ")=" << (*Ait).second << ";\n"; 
  }
}

void MapMatrixFactory:: computeRhs(ScalarFunction& func){ 
    rhs->clear(); 
    idof->clear(); 
    jdof->clear(); 
    int element = -1; 

    if (mesh->numSpaceDim() == 2) {

        double p0[2]; 
        double p1[2]; 
        double p2[2]; 

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

	    ++element;

      	    // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
      
            // Create double arrays with the data from the vertices
            p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
            p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
            p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

      
	    switch(order1) {
	      case 1 :  rhs_factory_2D_order1(*rhs,*idof,func,element,p0,p1,p2); break; 
	      case 2 :  rhs_factory_2D_order2(*rhs,*idof,func,element,p0,p1,p2); break; 
	      case 3 :  rhs_factory_2D_order3(*rhs,*idof,func,element,p0,p1,p2); break; 
	      case 4 :  rhs_factory_2D_order4(*rhs,*idof,func,element,p0,p1,p2); break; 
	      case 5 :  rhs_factory_2D_order5(*rhs,*idof,func,element,p0,p1,p2); break; 
	    }
	}
    }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);
      

      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


      switch(order1) {
//	      case 1 :  rhs_factory_3D_order1(*rhs,*idof,element,p0,p1,p2,p3); break; 
//	      case 2 :  rhs_factory_3D_order2(*rhs,*idof,element,p0,p1,p2,p3); break; 
//	      case 3 :  rhs_factory_3D_order3(*rhs,*idof,element,p0,p1,p2,p3); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeMassMatrix(bool append_to_matrix){ 

    if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
    }
    int element = -1; 

    if (mesh->numSpaceDim() == 2) {

        double p0[2]; 
        double p1[2]; 
        double p2[2]; 

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

	    ++element;

      	    // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
      
            // Create double arrays with the data from the vertices
           p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
           p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
           p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

      
	    switch(order1) {
	      case 1 :  matrix_factory_mass_2D_order1(*matrix,*idof,element,p0,p1,p2); break; 
	      case 2 :  matrix_factory_mass_2D_order2(*matrix,*idof,element,p0,p1,p2); break; 
	      case 3 :  matrix_factory_mass_2D_order3(*matrix,*idof,element,p0,p1,p2); break; 
	      case 4 :  matrix_factory_mass_2D_order4(*matrix,*idof,element,p0,p1,p2); break; 
	      case 5 :  matrix_factory_mass_2D_order5(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 6 :  matrix_factory_mass_2D_order6(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 7 :  matrix_factory_mass_2D_order7(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 8 :  matrix_factory_mass_2D_order8(*matrix,*idof,element,p0,p1,p2); break;  
//	      case 9 :  matrix_factory_mass_2D_order9(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 10 : matrix_factory_mass_2D_order10(*matrix,*dof,element,p0,p1,p2); break; 
//
	    }


	}
      
    }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);
      

      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


      switch(order1) {
	      case 1 :  matrix_factory_mass_3D_order1(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 2 :  matrix_factory_mass_3D_order2(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 3 :  matrix_factory_mass_3D_order3(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 4 :  matrix_factory_mass_3D_order4(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 5 :  matrix_factory_mass_3D_order5(*matrix,*idof,element,p0,p1,p2,p3); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeMassMatrix(ScalarFunction& func, bool append_to_matrix){ 

    if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
    }

    int element = -1; 
    double c = 0.0; 

    if (mesh->numSpaceDim() == 2) {

        double p0[2]; 
        double p1[2]; 
        double p2[2]; 
        double midpoint[2]; 
        ptv pm(2);

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

	    ++element;

      	    // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
      
            // Create double arrays with the data from the vertices
           p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
           p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
           p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

           midpoint[0] = (p0[0] + p1[0] + p2[0])/3; 
           midpoint[1] = (p0[1] + p1[1] + p2[1])/3; 

           pm.redim(2,midpoint); 
           c = func(pm);  


      
	    switch(order1) {
	      case 1 :  matrix_factory_mass_2D_with_c_order1(*matrix,*idof,element,p0,p1,p2,c); break; 
	      case 2 :  matrix_factory_mass_2D_with_c_order2(*matrix,*idof,element,p0,p1,p2,c); break; 
	      case 3 :  matrix_factory_mass_2D_with_c_order3(*matrix,*idof,element,p0,p1,p2,c); break; 
	      case 4 :  matrix_factory_mass_2D_with_c_order4(*matrix,*idof,element,p0,p1,p2,c); break; 
	      case 5 :  matrix_factory_mass_2D_with_c_order5(*matrix,*idof,element,p0,p1,p2,c); break; 
//	      case 6 :  matrix_factory_mass_2D_order6(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 7 :  matrix_factory_mass_2D_order7(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 8 :  matrix_factory_mass_2D_order8(*matrix,*idof,element,p0,p1,p2); break;  
//	      case 9 :  matrix_factory_mass_2D_order9(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 10 : matrix_factory_mass_2D_order10(*matrix,*dof,element,p0,p1,p2); break; 
//
	    }

	}
      
    }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 
    double midpoint[3]; 
    ptv pm(3);

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);
      

      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 

      midpoint[0] = (p0[0] + p1[0] + p2[0] + p3[0])/4; 
      midpoint[1] = (p0[1] + p1[1] + p2[1] + p3[1])/4; 
      midpoint[2] = (p0[2] + p1[2] + p2[2] + p3[2])/4; 

      pm.redim(3,midpoint); 
      c = func(pm);  


      switch(order1) {
	      case 1 :  matrix_factory_mass_3D_with_c_order1(*matrix,*idof,element,p0,p1,p2,p3,c); break; 
	      case 2 :  matrix_factory_mass_3D_with_c_order2(*matrix,*idof,element,p0,p1,p2,p3,c); break; 
	      case 3 :  matrix_factory_mass_3D_with_c_order3(*matrix,*idof,element,p0,p1,p2,p3,c); break; 
//	      case 4 :  matrix_factory_mass_3D_with_c_order4(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 5 :  matrix_factory_mass_3D_with_c_order5(*matrix,*idof,element,p0,p1,p2,p3); break; 
      }
    }
  }
}


void MapMatrixFactory:: computeVectorMassMatrix(bool append_to_matrix){ 

    if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
    }

    int element = -1; 

    if (mesh->numSpaceDim() == 2) {

        double p0[2]; 
        double p1[2]; 
        double p2[2]; 

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

	    ++element;

      	    // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
      
            // Create double arrays with the data from the vertices
            p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
            p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
            p2[0] = v2.coord().x;  p2[1] = v2.coord().y;
      
	    switch(order1) {
	      case 1 :  matrix_factory_vector_mass_2D_order1(*matrix,*idof,element,p0,p1,p2); break; 
	      case 2 :  matrix_factory_vector_mass_2D_order2(*matrix,*idof,element,p0,p1,p2); break; 
	      case 3 :  matrix_factory_vector_mass_2D_order3(*matrix,*idof,element,p0,p1,p2); break; 
	      case 4 :  matrix_factory_vector_mass_2D_order4(*matrix,*idof,element,p0,p1,p2); break; 
	      case 5 :  matrix_factory_vector_mass_2D_order5(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 6 :  matrix_factory_vector_mass_2D_order6(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 7 :  matrix_factory_vector_mass_2D_order7(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 8 :  matrix_factory_vector_mass_2D_order8(*matrix,*idof,element,p0,p1,p2); break;  
//	      case 9 :  matrix_factory_vector_mass_2D_order9(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 10 : matrix_factory_vector_mass_2D_order10(*matrix,*idof,element,p0,p1,p2); break; 
	    }
	}
      
    }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);
      

      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


      switch(order1) {
	      case 1 :  matrix_factory_vector_mass_3D_order1(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 2 :  matrix_factory_vector_mass_3D_order2(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 3 :  matrix_factory_vector_mass_3D_order3(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 4 :  matrix_factory_vector_mass_3D_order4(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 5 :  matrix_factory_vector_mass_3D_order5(*matrix,*idof,element,p0,p1,p2,p3); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeStiffnessMatrix(bool append_to_matrix){ 

  if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
  }

  int element = -1; 

  if (mesh->numSpaceDim() == 2) {

    double p0[2]; 
    double p1[2]; 
    double p2[2]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y;


      switch(order1) {
	      case 1 :  matrix_factory_stiffness_2D_order1(*matrix,*idof,element,p0,p1,p2); break; 
	      case 2 :  matrix_factory_stiffness_2D_order2(*matrix,*idof,element,p0,p1,p2); break; 
	      case 3 :  matrix_factory_stiffness_2D_order3(*matrix,*idof,element,p0,p1,p2); break; 
	      case 4 :  matrix_factory_stiffness_2D_order4(*matrix,*idof,element,p0,p1,p2); break; 
	      case 5 :  matrix_factory_stiffness_2D_order5(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 6 :  matrix_factory_stiffness_2D_order6(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 7 :  matrix_factory_stiffness_2D_order7(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 8 :  matrix_factory_stiffness_2D_order8(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 9 :  matrix_factory_stiffness_2D_order9(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 10 : matrix_factory_stiffness_2D_order10(*matrix,*idof,element,p0,p1,p2); break; 
      }
    }
  }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 

      switch(order1) {
	      case 1 :  matrix_factory_stiffness_3D_order1(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 2 :  matrix_factory_stiffness_3D_order2(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 3 :  matrix_factory_stiffness_3D_order3(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 4 :  matrix_factory_stiffness_3D_order4(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 5 :  matrix_factory_stiffness_3D_order5(*matrix,*idof,element,p0,p1,p2,p3); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeStiffnessMatrix(TensorFunction& tf, bool append_to_matrix){ 

  if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
  }

  int element = -1; 

  if (mesh->numSpaceDim() == 2) {

    double M[2][2]; // conductivity tensor 
    double p0[2]; 
    double p1[2]; 
    double p2[2]; 
    double midpoint[2]; 
    ptv pm(2);

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element; 

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

      // compute conductivity   
      midpoint[0] = (p0[0] + p1[0] + p2[0])/3; 
      midpoint[1] = (p0[1] + p1[1] + p2[1])/3; 

      pm.redim(2,midpoint); 
      tf(pm, &M[0][0], element);
     

      switch(order1) {
	      case 1 :  matrix_factory_stiffness_2D_with_M_order1(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 2 :  matrix_factory_stiffness_2D_with_M_order2(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 3 :  matrix_factory_stiffness_2D_with_M_order3(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 4 :  matrix_factory_stiffness_2D_with_M_order4(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 5 :  matrix_factory_stiffness_2D_with_M_order5(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 6 :  matrix_factory_stiffness_2D_with_M_order6(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 7 :  matrix_factory_stiffness_2D_with_M_order7(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 8 :  matrix_factory_stiffness_2D_with_M_order8(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 9 :  matrix_factory_stiffness_2D_with_M_order9(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 10 : matrix_factory_stiffness_2D_with_M_order10(*matrix,*idof,element,M,p0,p1,p2); break; 
      }
    }
  }
  else if (mesh->numSpaceDim() == 3) {

    double M[3][3]; //conductivity tensor
    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 
    double midpoint[3]; 
    ptv pm(3);

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 

      // compute conductivity   
      midpoint[0] = (p0[0] + p1[0] + p2[0] + p3[0])/4; 
      midpoint[1] = (p0[1] + p1[1] + p2[1] + p3[1])/4; 
      midpoint[2] = (p0[2] + p1[2] + p2[2] + p3[2])/4; 

      pm.redim(3,midpoint); 
      tf(pm, &M[0][0], element);

      switch(order1) {
	      case 1 :  matrix_factory_stiffness_3D_with_M_order1(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
	      case 2 :  matrix_factory_stiffness_3D_with_M_order2(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
	      case 3 :  matrix_factory_stiffness_3D_with_M_order3(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
//	      case 4 :  matrix_factory_stiffness_3D_with_M_order4(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
//	      case 5 :  matrix_factory_stiffness_3D_with_M_order5(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeConvectionMatrix(VectorFunction& func, bool append_to_matrix){ 

  if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
  }

  int element = -1; 

  if (mesh->numSpaceDim() == 2) {

    double p0[2]; 
    double p1[2]; 
    double p2[2]; 
    double midpoint[2]; 
    ptv pm(2);

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

      midpoint[0] = (p0[0] + p1[0] + p2[0])/3; 
      midpoint[1] = (p0[1] + p1[1] + p2[1])/3; 

      pm.redim(2,midpoint); 
      ptv vel(2); 
      func(pm,vel);  
     
      double vel0 = vel[0];  
      double vel1 = vel[1];  

      switch(order1) {
	      case 1 :  matrix_factory_conv_2D_order1(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
	      case 2 :  matrix_factory_conv_2D_order2(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
	      case 3 :  matrix_factory_conv_2D_order3(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
	      case 4 :  matrix_factory_conv_2D_order4(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
	      case 5 :  matrix_factory_conv_2D_order5(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
//	      case 6 :  matrix_factory_conv_2D_order6(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
//	      case 7 :  matrix_factory_conv_2D_order7(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
//	      case 8 :  matrix_factory_conv_2D_order8(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
//	      case 9 :  matrix_factory_conv_2D_order9(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
//	      case 10 : matrix_factory_conv_2D_order10(*matrix,*idof,element,p0,p1,p2,vel0,vel1); break; 
      }
    }
  }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 
    double midpoint[3]; 
    ptv pm(3);
    ptv vel(3); 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 

      midpoint[0] = (p0[0] + p1[0] + p2[0] + p3[0])/4; 
      midpoint[1] = (p0[1] + p1[1] + p2[1] + p3[1])/4; 
      midpoint[2] = (p0[2] + p1[2] + p2[2] + p3[2])/4; 

      pm.redim(3,midpoint); 
      func(pm,vel);  
      double vel0 = vel[0];  
      double vel1 = vel[1];  
      double vel2 = vel[2];  

      switch(order1) {
	      case 1 :  matrix_factory_conv_3D_order1(*matrix,*idof,element,p0,p1,p2,p3,vel0,vel1,vel2); break; 
	      case 2 :  matrix_factory_conv_3D_order2(*matrix,*idof,element,p0,p1,p2,p3,vel0,vel1,vel2); break; 
	      case 3 :  matrix_factory_conv_3D_order3(*matrix,*idof,element,p0,p1,p2,p3,vel0,vel1,vel2); break; 
//	      case 4 :  matrix_factory_conv_3D_order4(*matrix,*idof,element,p0,p1,p2,p3,vel0,vel1,vel2); break; 
//	      case 5 :  matrix_factory_conv_3D_order5(*matrix,*idof,element,p0,p1,p2,p3,vel0,vel1,vel2); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeVectorConvectionMatrix(double* vel, int size, bool append_to_matrix){ 

    if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
    }
    int element = -1; 


    if (mesh->numSpaceDim() == 2) {

        double p0[2]; 
        double p1[2]; 
        double p2[2]; 

        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

	    ++element;

      	    // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
      
            // Create double arrays with the data from the vertices
            p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
            p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
            p2[0] = v2.coord().x;  p2[1] = v2.coord().y;
      
	    switch(order1) {
	      case 1 :  matrix_factory_vector_conv_2D_order1(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
	      case 2 :  matrix_factory_vector_conv_2D_order2(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
	      case 3 :  matrix_factory_vector_conv_2D_order3(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
	      case 4 :  matrix_factory_vector_conv_2D_order4(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
	      case 5 :  matrix_factory_vector_conv_2D_order5(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
//	      case 6 :  matrix_factory_vector_conv_2D_order6(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
//	      case 7 :  matrix_factory_vector_conv_2D_order7(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
//	      case 8 :  matrix_factory_vector_conv_2D_order8(*matrix,*idof,element,p0,p1,p2,vel,size); break;  
//	      case 9 :  matrix_factory_vector_conv_2D_order9(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
//	      case 10 : matrix_factory_vector_conv_2D_order10(*matrix,*idof,element,p0,p1,p2,vel,size); break; 
	    }
	}
      
    }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);
      

      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


      switch(order1) {
	      case 1 :  matrix_factory_vector_conv_3D_order1(*matrix,*idof,element,p0,p1,p2,p3,vel,size); break; 
	      case 2 :  matrix_factory_vector_conv_3D_order2(*matrix,*idof,element,p0,p1,p2,p3,vel,size); break; 
	      case 3 :  matrix_factory_vector_conv_3D_order3(*matrix,*idof,element,p0,p1,p2,p3,vel,size); break; 
//	      case 4 :  matrix_factory_vector_conv_3D_order4(*matrix,*idof,element,p0,p1,p2,p3,vel,size); break; 
//	      case 5 :  matrix_factory_vector_conv_3D_order5(*matrix,*idof,element,p0,p1,p2,p3,vel,size); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeVectorStiffnessMatrix(bool append_to_matrix){ 

  if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
  }
  int element = -1; 


  if (mesh->numSpaceDim() == 2) {

    double p0[2]; 
    double p1[2]; 
    double p2[2]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y;


      switch(order1) {
	      case 1 :  matrix_factory_vector_stiffness_2D_order1(*matrix,*idof,element,p0,p1,p2); break; 
	      case 2 :  matrix_factory_vector_stiffness_2D_order2(*matrix,*idof,element,p0,p1,p2); break; 
	      case 3 :  matrix_factory_vector_stiffness_2D_order3(*matrix,*idof,element,p0,p1,p2); break; 
	      case 4 :  matrix_factory_vector_stiffness_2D_order4(*matrix,*idof,element,p0,p1,p2); break; 
	      case 5 :  matrix_factory_vector_stiffness_2D_order5(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 6 :  matrix_factory_vector_stiffness_2D_order6(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 7 :  matrix_factory_vector_stiffness_2D_order7(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 8 :  matrix_factory_vector_stiffness_2D_order8(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 9 :  matrix_factory_vector_stiffness_2D_order9(*matrix,*idof,element,p0,p1,p2); break; 
//	      case 10 : matrix_factory_vector_stiffness_2D_order10(*matrix,*idof,element,p0,p1,p2); break; 
      }
    }
  }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 

      switch(order1) {
	      case 1 :  matrix_factory_vector_stiffness_3D_order1(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 2 :  matrix_factory_vector_stiffness_3D_order2(*matrix,*idof,element,p0,p1,p2,p3); break; 
	      case 3 :  matrix_factory_vector_stiffness_3D_order3(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 4 :  matrix_factory_vector_stiffness_3D_order4(*matrix,*idof,element,p0,p1,p2,p3); break; 
//	      case 5 :  matrix_factory_vector_stiffness_3D_order5(*matrix,*idof,element,p0,p1,p2,p3); break; 
      }
    }
  }
}


void MapMatrixFactory:: computeVectorStiffnessMatrix(TensorFunction& tf, bool append_to_matrix){ 

  if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
  }
  int element = -1; 


  if (mesh->numSpaceDim() == 2) {

    double M[2][2]; // conductivity tensor 
    double p0[2]; 
    double p1[2]; 
    double p2[2]; 
    double midpoint[2]; 
    ptv pm(2);

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

      // compute conductivity   
      midpoint[0] = (p0[0] + p1[0] + p2[0])/3; 
      midpoint[1] = (p0[1] + p1[1] + p2[1])/3; 

      pm.redim(2,midpoint); 

      tf(pm, &M[0][0], (*cell).id());
     

      switch(order1) {
	      case 1 :  matrix_factory_vector_stiffness_2D_with_M_order1(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 2 :  matrix_factory_vector_stiffness_2D_with_M_order2(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 3 :  matrix_factory_vector_stiffness_2D_with_M_order3(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 4 :  matrix_factory_vector_stiffness_2D_with_M_order4(*matrix,*idof,element,M,p0,p1,p2); break; 
	      case 5 :  matrix_factory_vector_stiffness_2D_with_M_order5(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 6 :  matrix_factory_vector_stiffness_2D_with_M_order6(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 7 :  matrix_factory_vector_stiffness_2D_with_M_order7(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 8 :  matrix_factory_vector_stiffness_2D_with_M_order8(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 9 :  matrix_factory_vector_stiffness_2D_with_M_order9(*matrix,*idof,element,M,p0,p1,p2); break; 
//	      case 10 : matrix_factory_vector_stiffness_2D_with_M_order10(*matrix,*idof,element,M,p0,p1,p2); break; 
      }
    }
  }
  else if (mesh->numSpaceDim() == 3) {

    double M[3][3]; //conductivity tensor
    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 
    double midpoint[3]; 
    ptv pm(3);

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 

      // compute conductivity   
      midpoint[0] = (p0[0] + p1[0] + p2[0] + p3[0])/4; 
      midpoint[1] = (p0[1] + p1[1] + p2[1] + p3[1])/4; 
      midpoint[2] = (p0[2] + p1[2] + p2[2] + p3[2])/4; 

      pm.redim(3, midpoint); 
      tf(pm, &M[0][0], element);

      switch(order1) {
	      case 1 :  matrix_factory_vector_stiffness_3D_with_M_order1(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
	      case 2 :  matrix_factory_vector_stiffness_3D_with_M_order2(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
	      case 3 :  matrix_factory_vector_stiffness_3D_with_M_order3(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
//	      case 4 :  matrix_factory_vector_stiffness_3D_with_M_order4(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
//	      case 5 :  matrix_factory_vector_stiffness_3D_with_M_order5(*matrix,*idof,element,M,p0,p1,p2,p3); break; 
      }
    }
  }
}

void MapMatrixFactory:: computeDivMatrix(bool append_to_matrix){ 

  if (! append_to_matrix) {
      matrix->clear();
      idof->clear(); 
      jdof->clear(); 
  }

  int element = -1; 

  if (mesh->numSpaceDim() == 2) {

    double p0[2]; 
    double p1[2]; 
    double p2[2]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

      // Taylor-Hood
      if      ( order1 == 2 && order2 == 1) matrix_factory_div_2D_order2_1(*matrix,*idof,*jdof,element,p0,p1,p2); 
      // P2-P0
      else if ( order1 == 2 && order2 == 0) matrix_factory_div_2D_order2_1(*matrix,*idof,*jdof,element,p0,p1,p2); 
      // P3-P1
      else if ( order1 == 3 && order2 == 1) matrix_factory_div_2D_order3_1(*matrix,*idof,*jdof,element,p0,p1,p2); 
    }
  }
  else if (mesh->numSpaceDim() == 3) {

    double p0[3]; 
    double p1[3]; 
    double p2[3]; 
    double p3[3]; 

    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);


      // Create double arrays with the data from the vertices
      p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


      // Taylor-Hood
      if      ( order1 == 2 && order2 == 1) matrix_factory_div_3D_order2_1(*matrix,*idof,*jdof,element,p0,p1,p2,p3); 
      // P2-P0
//      else if ( order1 == 2 && order2 == 0) matrix_factory_div_3D_order2_1(*matrix,*idof,*jdof,element,p0,p1,p2,p3); 
      // P3-P1
      else if ( order1 == 3 && order2 == 1) matrix_factory_div_3D_order3_1(*matrix,*idof,*jdof,element,p0,p1,p2,p3); 

    }
  }
}

bool eq(double a, double b, double tolerance) {
  if (fabs(a-b) <= tolerance) return true; 
  else return false; 
}

void printDofs(Dof_ptv& dof, bool glob2loc ){ 
  for (int i=0; i < dof.size(); i++) {
    std::cout <<"dof "<<i<<" "<<dof.glob_dof(i)<<std::endl; 
    if ( glob2loc) {
      vector<pair<int,int> > dof_vec; 
      dof_vec = dof.glob2loc(i); 
      vector<pair<int,int> >::const_iterator iter; 
      for (iter = dof_vec.begin(); iter != dof_vec.end() ; iter++) {  
    	int e = (*iter).first;
    	int j = (*iter).second; 
    	std::cout <<"local e "<<e<<" local j "<<j<<" global i "<<dof.glob_dof(e,j)<<std::endl; 
      }
    }
  }
}

#ifdef HAVE_ML_EPETRA

//-- Epetra --------------------------------------------------------------------------

EpetraMatrixFactory:: EpetraMatrixFactory(Mesh* mesh_, Epetra_Comm* Comm_){
  mapfac = new MapMatrixFactory(mesh_); 
  Comm = Comm_; 
  Map = new Epetra_Map(1, 0, *Comm); 
  order1 = 1; 
  order2 = 1; 
}


Epetra_CrsMatrix* EpetraMatrixFactory:: computeMassMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeMassMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeMassMatrix(ScalarFunction& func, bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeMassMatrix(func,append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}



Epetra_CrsMatrix* EpetraMatrixFactory:: computeStiffnessMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeStiffnessMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeStiffnessMatrix(TensorFunction& tf, bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeStiffnessMatrix(tf,append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}


Epetra_CrsMatrix* EpetraMatrixFactory:: computeConvectionMatrix(VectorFunction& func,bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeConvectionMatrix(func,append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeVectorMassMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorMassMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeVectorStiffnessMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorStiffnessMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeVectorStiffnessMatrix(TensorFunction& tf, bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorStiffnessMatrix(tf,append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeVectorConvectionMatrix(double* v, int size,bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->computeVectorConvectionMatrix(v,size,append_to_matrix); 
  idof = mapfac->idof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}

Epetra_CrsMatrix* EpetraMatrixFactory:: computeDivMatrix(bool append_to_matrix) {
  mapfac->order1 = order1; 
  mapfac->order2 = order2; 
  mapfac->computeDivMatrix(append_to_matrix); 
  idof = mapfac->idof; 
  jdof = mapfac->jdof; 
  delete Map; 
  Map = new Epetra_Map(idof->size(), 0, *Comm); 
  return map2epetra(*(mapfac->matrix),*Map);  
}



Epetra_CrsMatrix* map2epetra(map<pair<int,int>,double>& A_from, Epetra_Map& emap) { 

  std::map<std::pair<int,int>,double>::const_iterator Ait;

  Epetra_CrsMatrix* A_to = new Epetra_FECrsMatrix(Copy, emap,1); 


//  dump_to_matlab(A_from, "A", "A.m"); 

  double value; 
  vector<int> j_at_row; 
  vector<double> value_at_row; 
  Ait=A_from.begin();
  int curr_row= (*Ait).first.first;  
  int i, j; 
  int size; 

  while ( Ait != A_from.end() ) { 

    i = (*Ait).first.first;
    j = (*Ait).first.second;
    value = (*Ait).second; 

    if ( i == curr_row) { 

      j_at_row.push_back(j); 
      value_at_row.push_back(value); 

    } else { // new row! 


      // 1. Fill Epetra_CrsMatrix.
      if ( j_at_row.size() != value_at_row.size() ) { 
         //FIXME shold this be an exception ?? 
         std::cout << "This is most likely a bug in the function map2epetra" <<std::endl; 
      }

      size = j_at_row.size(); 

      // create the temporal pointer arrays Indices and Values 
      int* Indices =  new int[size]; 
      double* Values = new double[size];

      // fill Indices and Values 
      for (int d=0; d < size; d++) { 
        Indices[d] = j_at_row[d]; 
        Values[d]  = value_at_row[d];  
      }

      // insert previous row into Epetra_CrsMatrix
      A_to->InsertGlobalValues (curr_row, size, Values, Indices);

      
      // 2. Clear the structures i_at_row, j_at_row and value_at_row.
      j_at_row.clear(); 
      value_at_row.clear(); 

      // delete temporal pointer arrays 
      delete [] Indices; 
      delete [] Values;

      // 3. Insert the new values.
      j_at_row.push_back(j); 
      value_at_row.push_back(value); 

      curr_row = i; 
    }
    Ait++; 
  }

  // Finally, fill the last values 
  size = j_at_row.size(); 

  // create the temporal pointer arrays Indices and Values 
  int* Indices =  new int[size]; 
  double* Values = new double[size];

  // fill Indices and Values 
  for (int d=0; d < size; d++) { 
    Indices[d] = j_at_row[d]; 
    Values[d]  = value_at_row[d];  
  }

  // insert into Epetra_CrsMatrix
  A_to->InsertGlobalValues (curr_row, size, Values, Indices);

      
  // 2. Clear the structures i_at_row, j_at_row and value_at_row.
  j_at_row.clear(); 
  value_at_row.clear(); 

  // delete temporal pointer arrays 
  delete [] Indices; 
  delete [] Values;

  // 3. Insert the new values.
  j_at_row.push_back(j); 
  value_at_row.push_back(value); 

  A_to->FillComplete(); 

  return A_to; 
  
}

#endif //HAVE_ML_EPETRA 

