#ifndef MatrixFactory_h_IS_INCLUDED
#define MatrixFactory_h_IS_INCLUDED

#include <dolfin.h>
#include <utility>
#include <map>
#include <iostream>
#include <fstream>
#include <TensorFunction.h>
#include <FastMatSparse.h>
#include <ptv.h>
#include <DofT.h>
#include <ScalarFunction.h>
#include <VectorFunction.h>


#include <ml_config.h>
#ifdef HAVE_ML_EPETRA 
#ifdef HAVE_MPI
#include "mpi.h"
#include "Epetra_MpiComm.h"
#else
#include "Epetra_SerialComm.h"
#endif
#include <Epetra_Map.h>
#include <Epetra_FECrsMatrix.h>
#include <Epetra_ConfigDefs.h>
#endif // HAVE_ML_EPETRA 




using namespace dolfin;
using namespace std;

typedef DofT<ptv,ptv_is_less> Dof_ptv; 

// forward declaration
class MapMatrixFactory; 

class MatrixFactory 
{
    public:
	MapMatrixFactory* mapfac; 
    Mesh *mesh;
	Dof_ptv  *idof; 
	Dof_ptv  *jdof; 
	int order1; 
	int order2; 

        MatrixFactory(Mesh *_mesh);

	// scalar mass matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeMassMatrix(bool append_to_matrix=false);
	// scalar mass matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeMassMatrix(ScalarFunction& func, bool append_to_matrix=false);
	// vector mass matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeVectorMassMatrix(bool append_to_matrix=false);


	// scalar stiffness matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeStiffnessMatrix(bool append_to_matrix=false); 
	// scalar stiffness matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeStiffnessMatrix(TensorFunction& tf, bool append_to_matrix=false); 
	// vector stiffness matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeVectorStiffnessMatrix(bool append_to_matrix=false); 
	// vector stiffness matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeVectorStiffnessMatrix(TensorFunction& tf, bool append_to_matrix=false); 

	// scalar convection matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeConvectionMatrix(VectorFunction& func, bool append_to_matrix=false); 
	// vector convection matrix based on Lagrangian elements with order order1  
        FastMatSparse*  computeVectorConvectionMatrix(double* vel, int size, bool append_to_matrix=false); 
	// vector div matrix based on vector Lagrangian elements with order order1  
	// and scalar Lagrangian elements with order2 
        FastMatSparse*  computeDivMatrix(bool append_to_matrix=false); 



};





class MapMatrixFactory 
{
    public:
        Mesh *mesh;
	Dof_ptv  *idof; 
	Dof_ptv  *jdof; 
        MapMatrixFactory(Mesh *_mesh);
        map<pair<int,int>,double> *matrix;
	map<int,double>            *rhs; 
	int order1; 
	int order2; 

	// scalar mass matrix based on Lagrangian elements with order order1  
        void computeMassMatrix(bool append_to_matrix=false);
	// scalar mass matrix based on Lagrangian elements with order order1  
        void computeMassMatrix(ScalarFunction& func, bool append_to_matrix=false);
	// vector mass matrix based on Lagrangian elements with order order1  
        void computeVectorMassMatrix(bool append_to_matrix=false);


	// scalar stiffness matrix based on Lagrangian elements with order order1  
        void computeStiffnessMatrix(bool append_to_matrix=false); 
	// scalar stiffness matrix based on Lagrangian elements with order order1  
        void computeStiffnessMatrix(TensorFunction& tf, bool append_to_matrix=false); 
	// vector stiffness matrix based on Lagrangian elements with order order1  
        void computeVectorStiffnessMatrix(bool append_to_matrix=false); 
	// vector stiffness matrix based on Lagrangian elements with order order1  
        void computeVectorStiffnessMatrix(TensorFunction& tf,bool append_to_matrix=false); 


	// scalar convection matrix based on Lagrangian elements with order order1  
        void computeConvectionMatrix(VectorFunction& func,bool append_to_matrix=false); 
	// vector convection matrix based on Lagrangian elements with order order1  
        void computeVectorConvectionMatrix(double* vel, int size,bool append_to_matrix=false); 

	// vector div matrix based on vector Lagrangian elements with order order1  
	// and scalar Lagrangian elements with order2 
        void computeDivMatrix(bool append_to_matrix=false); 

	void computeRhs(ScalarFunction& f);



};


void dump_to_matlab(map<pair<int,int>,double> A, const char *varname, const char *s);

bool eq(double a, double b, double tolerance);

void printDofs(Dof_ptv& dof, bool glob2loc=false); 


#ifdef HAVE_ML_EPETRA


class EpetraMatrixFactory 
{
    public:
  
	Epetra_Comm       *Comm; 
	Epetra_Map        *Map; 
	MapMatrixFactory  *mapfac; 
        Mesh              *mesh;
	Dof_ptv           *idof; 
	Dof_ptv           *jdof; 
	int order1; 
	int order2; 

        EpetraMatrixFactory(Mesh *_mesh, Epetra_Comm* Comm);

	// scalar mass matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeMassMatrix(bool append_to_matrix=false);
	// scalar mass matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeMassMatrix(ScalarFunction& func, bool append_to_matrix=false);
	// vector mass matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeVectorMassMatrix(bool append_to_matrix=false);


	// scalar stiffness matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeStiffnessMatrix(bool append_to_matrix=false); 
	// scalar stiffness matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeStiffnessMatrix(TensorFunction& tf, bool append_to_matrix=false); 
	// vector stiffness matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeVectorStiffnessMatrix(bool append_to_matrix=false); 
	// vector stiffness matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeVectorStiffnessMatrix(TensorFunction& tf, bool append_to_matrix=false); 

	// scalar convection matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeConvectionMatrix(VectorFunction& func,bool append_to_matrix=false); 
	// vector convection matrix based on Lagrangian elements with order order1  
        Epetra_CrsMatrix*  computeVectorConvectionMatrix(double* vel, int size,bool append_to_matrix=false); 

	// vector div matrix based on vector Lagrangian elements with order order1  
	// and scalar Lagrangian elements with order2 
        Epetra_CrsMatrix*  computeDivMatrix(bool append_to_matrix=false); 


};


Epetra_CrsMatrix* map2epetra(map<pair<int,int>,double>& A_from, Epetra_Map& emap);

#endif // HAVE_ML_EPETRA


#endif
