#include <MatrixFactory.h>
#include <ElementMatrixCalculator.h>
#include <compute_conductivity.h>
#include <ptv.h>

using namespace dolfin;
using namespace std;

MatrixFactory:: MatrixFactory(Mesh *_mesh) {
    mesh = _mesh;
    matrix = new map<pair<int,int>,double>();
    tf = new ConstTensorFunction(mesh->numSpaceDim());
}

void MatrixFactory:: dump_to_matlab(std::map<pair<int,int>,double> A, const char *varname, const char *s)
{
  ofstream outfile(s, ios::out);
  outfile << "%The matrix\n"; 
  std::map<std::pair<int,int>,double>::iterator Ait;
  for(Ait=A.begin(); Ait!=A.end(); Ait++){
    outfile << varname << "(" << (*Ait).first.first +1 << ","
            << (*Ait).first.second +1 << ")=" << (*Ait).second << ";\n"; 
  }
}


FastMatSparse* MatrixFactory:: computeMassMatrix(){ 
    matrix->clear();
    if (mesh->numSpaceDim() == 2) {
        double A[3][3];
        for (CellIterator cell(*mesh); !cell.end(); ++cell) {

	
      	    // Obtain vertices from dolfin mesh
            Vertex& v0 = (*cell).vertex(0);
            Vertex& v1 = (*cell).vertex(1);
            Vertex& v2 = (*cell).vertex(2);
      
            // Obtain corresponding dofs in global matrix
            int  ig0 = (*cell).vertex(0).id();
            int  ig1 = (*cell).vertex(1).id();
            int  ig2 = (*cell).vertex(2).id();
      
            // Create double arrays with the data from the vertices
            double p0[2]; p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
            double p1[2]; p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
            double p2[2]; p2[0] = v2.coord().x;  p2[1] = v2.coord().y;
      
            // compute element matrix 
            matrix_factory_mass_2D(A,p0,p1,p2);
      
            // put element matrix in global matrix
            pair<int,int> index; 
            index.first = ig0; index.second = ig0;
            (*matrix)[index] += A[0][0];
            index.second = ig1;
            (*matrix)[index] += A[0][1];
            index.second = ig2;
            (*matrix)[index] += A[0][2];
      
      
            index.first = ig1; index.second = ig0;
            (*matrix)[index] += A[1][0];
            index.second = ig1;
            (*matrix)[index] += A[1][1];
            index.second = ig2;
            (*matrix)[index] += A[1][2];
      
            index.first = ig2; index.second = ig0;
            (*matrix)[index] += A[2][0];
            index.second = ig1;
            (*matrix)[index] += A[2][1];
            index.second = ig2;
            (*matrix)[index] += A[2][2];
        }
    }
  else if (mesh->numSpaceDim() == 3) {
    double A[4][4];
    double M[3][3]; 
    for (CellIterator cell(*mesh); !cell.end(); ++cell) {
	          // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);

      

      // Obtain corresponding dofs in global matrix
      int  ig0 = (*cell).vertex(0).id();
      int  ig1 = (*cell).vertex(1).id();
      int  ig2 = (*cell).vertex(2).id();
      int  ig3 = (*cell).vertex(3).id();

      // Create double arrays with the data from the vertices
      double p0[3]; p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      double p1[3]; p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      double p2[3]; p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      double p3[3]; p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 


      //compute element matrix 
      matrix_factory_mass_3D(A,p0,p1,p2,p3);

      // put element matrix in global matrix: 
      pair<int,int> index; 
      index.first = ig0; index.second = ig0;
      (*matrix)[index] += A[0][0];
      index.second = ig1;
      (*matrix)[index] += A[0][1];
      index.second = ig2;
      (*matrix)[index] += A[0][2];
      index.second = ig3;
      (*matrix)[index] += A[0][3];


      index.first = ig1; index.second = ig0;
      (*matrix)[index] += A[1][0];
      index.second = ig1;
      (*matrix)[index] += A[1][1];
      index.second = ig2;
      (*matrix)[index] += A[1][2];
      index.second = ig3;
      (*matrix)[index] += A[1][3];

      index.first = ig2; index.second = ig0;
      (*matrix)[index] += A[2][0];
      index.second = ig1;
      (*matrix)[index] += A[2][1];
      index.second = ig2;
      (*matrix)[index] += A[2][2];
      index.second = ig3;
      (*matrix)[index] += A[2][3];

      index.first = ig3; index.second = ig0;
      (*matrix)[index] += A[3][0];
      index.second = ig1;
      (*matrix)[index] += A[3][1];
      index.second = ig2;
      (*matrix)[index] += A[3][2];
      index.second = ig3;
      (*matrix)[index] += A[3][3];
    }
  }
  return new FastMatSparse(*matrix);
}

FastMatSparse* MatrixFactory:: computeStiffnessMatrix(){ 
    matrix->clear();
  if (mesh->numSpaceDim() == 2) {
    double A[3][3];
    double M[2][2]; 
    int element = -1;
    ptv pm(2);
    for (CellIterator cell(*mesh); !cell.end(); ++cell) {

      ++element;

      // Obtain vertices from dolfin mesh
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);

      // Obtain corresponding dofs in global matrix
      int  ig0 = (*cell).vertex(0).id();
      int  ig1 = (*cell).vertex(1).id();
      int  ig2 = (*cell).vertex(2).id();

      // Create double arrays with the data from the vertices
      double p0[2]; p0[0] = v0.coord().x;  p0[1] = v0.coord().y;
      double p1[2]; p1[0] = v1.coord().x;  p1[1] = v1.coord().y;
      double p2[2]; p2[0] = v2.coord().x;  p2[1] = v2.coord().y;

      // compute conductivity   
      double midpoint[2]; 
      midpoint[0] = (p0[0] + p1[0] + p2[0])/3; 
      midpoint[1] = (p0[1] + p1[1] + p2[1])/3; 

      pm.fill(midpoint);
      (*tf)(pm, &M[0][0], element);


      // compute element matrix 
      matrix_factory_stiffness_2D(A,M,p0,p1,p2);

      // put element matrix in global matrix
      pair<int,int> index; 
      index.first = ig0; index.second = ig0;
      (*matrix)[index] += A[0][0];
      index.second = ig1;
      (*matrix)[index] += A[0][1];
      index.second = ig2;
      (*matrix)[index] += A[0][2];


      index.first = ig1; index.second = ig0;
      (*matrix)[index] += A[1][0];
      index.second = ig1;
      (*matrix)[index] += A[1][1];
      index.second = ig2;
      (*matrix)[index] += A[1][2];

      index.first = ig2; index.second = ig0;
      (*matrix)[index] += A[2][0];
      index.second = ig1;
      (*matrix)[index] += A[2][1];
      index.second = ig2;
      (*matrix)[index] += A[2][2];
    }
  }
  else if (mesh->numSpaceDim() == 3) {
    double A[4][4];
    double M[3][3];
    int element = -1;
    ptv pm(3);
    for (CellIterator cell(*mesh); !cell.end(); ++cell) {
	          // Obtain vertices from dolfin mesh
      ++element;
      Vertex& v0 = (*cell).vertex(0);
      Vertex& v1 = (*cell).vertex(1);
      Vertex& v2 = (*cell).vertex(2);
      Vertex& v3 = (*cell).vertex(3);

      // Obtain corresponding dofs in global matrix
      int  ig0 = (*cell).vertex(0).id();
      int  ig1 = (*cell).vertex(1).id();
      int  ig2 = (*cell).vertex(2).id();
      int  ig3 = (*cell).vertex(3).id();

      // Create double arrays with the data from the vertices
      double p0[3]; p0[0] = v0.coord().x;  p0[1] = v0.coord().y; p0[2] = v0.coord().z; 
      double p1[3]; p1[0] = v1.coord().x;  p1[1] = v1.coord().y; p1[2] = v1.coord().z; 
      double p2[3]; p2[0] = v2.coord().x;  p2[1] = v2.coord().y; p2[2] = v2.coord().z; 
      double p3[3]; p3[0] = v3.coord().x;  p3[1] = v3.coord().y; p3[2] = v3.coord().z; 
      // compute conductivity   
      double midpoint[3]; 
      midpoint[0] = (p0[0] + p1[0] + p2[0] + p3[0])/4; 
      midpoint[1] = (p0[1] + p1[1] + p2[1] + p3[1])/4; 
      midpoint[2] = (p0[2] + p1[2] + p2[2] + p3[2])/4; 

      pm.fill(midpoint);
      (*tf)(pm, &M[0][0], element);


      //compute element matrix 
      matrix_factory_stiffness_3D(A,M,p0,p1,p2,p3);

      // put element matrix in global matrix: 
      pair<int,int> index; 
      index.first = ig0; index.second = ig0;
      (*matrix)[index] += A[0][0];
      index.second = ig1;
      (*matrix)[index] += A[0][1];
      index.second = ig2;
      (*matrix)[index] += A[0][2];
      index.second = ig3;
      (*matrix)[index] += A[0][3];


      index.first = ig1; index.second = ig0;
      (*matrix)[index] += A[1][0];
      index.second = ig1;
      (*matrix)[index] += A[1][1];
      index.second = ig2;
      (*matrix)[index] += A[1][2];
      index.second = ig3;
      (*matrix)[index] += A[1][3];

      index.first = ig2; index.second = ig0;
      (*matrix)[index] += A[2][0];
      index.second = ig1;
      (*matrix)[index] += A[2][1];
      index.second = ig2;
      (*matrix)[index] += A[2][2];
      index.second = ig3;
      (*matrix)[index] += A[2][3];

      index.first = ig3; index.second = ig0;
      (*matrix)[index] += A[3][0];
      index.second = ig1;
      (*matrix)[index] += A[3][1];
      index.second = ig2;
      (*matrix)[index] += A[3][2];
      index.second = ig3;
      (*matrix)[index] += A[3][3];
    }
  }

  return new FastMatSparse(*matrix);
}

/*
void MatrixFactory:: set_const_cond(int n, double* x)
{
    if (tf  <> NULL) delete tf;
    tf = new ConstTensorFunc();
}


void MatrixFactory:: set_interp_cond()
{
    cond2d_pt = &cond2D;
    cond3d_pt = &cond3D;
}

*/
