# -*- coding: utf-8 -*-


"""
@license: Copyright (C) 2005
Author: smund degrd, Ola Skavhaug

Simula Research Laboratory AS

This file is part of PyCC.

PyCC  free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

PyCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PyFDM; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


ConjGrad: A generic implementation of the (preconditioned) conjugate gradient method. 
"""


ConjGradError = "Error in ConjGrad"

import Numeric as _n #/numarray
#from Utils import _n,inner
from math import sqrt

def inner(u,v):
    """Compute innerproduct of u and v.
       It is not computed here, we just check type of operands and
       call a suitable innerproduct method"""

    if isinstance(u,_n.ArrayType):
        # assume both are numarrays:
        return _n.dot(u,v)
    else:
        # assume that left operand implements inner
        return u.inner(v)

def conjgrad(A, x, b, tolerance=1.0E-05, relativeconv=False, maxiter=300):
    """
    conjgrad(A,x,b): Solve Ax = b with the conjugate gradient method.

    @param A: Matrix or discrete operator. Must support A*x to operate on x.

    @param x: Anything that A can operate on, as well as support inner product
    and multiplication/addition/subtraction with scalar and something of the
    same type

    @param b: Right-hand side, probably the same type as x.

    @param tolerance: Convergence criterion
    @type tolerance: float

    @param relativeconv: Control relative vs. absolute convergence criterion.
    That is, if relativeconv is True, ||r||_2/||r_init||_2 is used as
    convergence criterion, else just ||r||_2 is used.  Actually ||r||_2^2 is
    used, since that save a few cycles.

    @type relativeconv: bool

    @return:  the solution x. 
    """

    r = b - A*x
    p = r.copy()
    r0 = inner(r,r) 
    
    if relativeconv:
        tolerance *= sqrt(inner(b,b))
    iter = 0
    # Alloc w
    while sqrt(r0) > tolerance and iter <= maxiter:
        print iter
        w = A*p
        # A.prod(p,w)
        a = r0/inner(p,w)
        x += a*p
        r -= a*w
        r1 = inner(r,r)
        p = r + (r1/r0)*p
        r0 = r1
        iter += 1
    #print "CG finished, iter: %d, ||e||=%e" % (iter,sqrt(r0))
    return x

def precondconjgrad(B, A, x, b, tolerance=1.0E-05, relativeconv=False):
    """
    conjgrad(B,A,x,b): Solve Ax = b with the preconditioned conjugate gradient method.

    @param B: Preconditioner supporting the __mul__ operator for operating on
    x.

    @param A: Matrix or discrete operator. Must support A*x to operate on x.

    @param x: Anything that A can operate on, as well as support inner product
    and multiplication/addition/subtraction with scalar and something of the
    same type

    @param b: Right-hand side, probably the same type as x.

    @param tolerance: Convergence criterion
    @type tolerance: float

    @param relativeconv: Control relative vs. absolute convergence criterion.
    That is, if relativeconv is True, ||r||_2/||r_init||_2 is used as
    convergence criterion, else just ||r||_2 is used.  Actually ||r||_2^2 is
    used, since that save a few cycles.

    @type relativeconv: bool

    @return:  the solution x. 

    """

    r0 = b - A*x
    r = r0.copy()
    z = r.copy()
#    t = r.copy()
    s = B*r
    p = s.copy()
    rho = rho0 = inner(s, r)
    #print "rho0 = " + str(rho)
    if relativeconv:
        tolerance *= sqrt(inner(B*b,b))
    iter = 0
    while sqrt(rho0) > tolerance:
        #print sqrt(rho0), tolerance, sqrt(rho0) > tolerance
        z = A*p
#        A.prod(p,z)
#        print z[:10]
        t = B*z
        #B.prod(z,t)
        g = inner(p,z)
        a = rho0/g
        x += a*p
        r -= a*z
        s -= a*t
        rho = inner(s, r)
        #print "rho = " + str(rho)
        b = rho/rho0
#        p = s + b*p
        p *= b; p += s
        rho0 = rho
        iter += 1
    #print "PCG finished, iter: %d, ||rho||=%e" % (iter,sqrt(rho0))
    return x
