#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <map>
#include <stdexcept>

#include <vec.h>
#include <ginac/numeric.h>
#include <ginac/lst.h>
#include <ginac/idx.h>
#include <ginac/indexed.h>
#include <ginac/add.h>
#include <ginac/power.h>
#include <ginac/symbol.h>
#include <ginac/operators.h>
#include <ginac/normal.h>
#include <ginac/archive.h>

#include <tools.h>


namespace GiNaC {

GINAC_IMPLEMENT_REGISTERED_CLASS_OPT(vec, basic,
  print_func<print_context>(&vec::do_print).
  print_func<print_latex>(&vec::do_print_latex).
  print_func<print_tree>(&vec::do_print_tree).
  print_func<print_python_repr>(&vec::do_print_python_repr))

//////////
// default constructor
//////////

/** Default ctor.  Initializes to 1 x 1-dimensional zero-vec. */
vec::vec() : inherited(TINFO_vec), size_(1), m(1, DUMMY)
{
	setflag(status_flags::not_shareable);
}

vec::vec(unsigned r)
  : inherited(TINFO_vec), size_(r), m(r, DUMMY)
{
	setflag(status_flags::not_shareable);
}

// protected

/** Ctor from representation, for internal use only. */
vec::vec(const exvector & m2)
  : inherited(TINFO_vec), m(m2)
{
	setflag(status_flags::not_shareable);
}

vec::vec(const lst & l) : inherited(TINFO_vec), m(size_, DUMMY)
{
	setflag(status_flags::not_shareable);

	size_t i = 0;
	for (lst::const_iterator it = l.begin(); it != l.end(); ++it, ++i) {
		m[i] = *it;
	}
}

//////////
// archiving
//////////

vec::vec(const archive_node &n, lst &sym_lst) : inherited(n, sym_lst)
{
	setflag(status_flags::not_shareable);

	if (!(n.find_unsigned("size", size)) )
		throw (std::runtime_error("unknown vec dimensions in archive"));
	m.reserve(size_);
	for (unsigned int i=0; true; i++) {
		ex e;
		if (n.find_ex("m", e, sym_lst, i))
			m.push_back(e);
		else
			break;
	}
}

void vec::archive(archive_node &n) const
{
	inherited::archive(n);
	n.add_unsigned("size", size_);
	exvector::const_iterator i = m.begin(), iend = m.end();
	while (i != iend) {
		n.add_ex("m", *i);
		++i;
	}
}

DEFAULT_UNARCHIVE(vec)

//////////
// functions overriding virtual functions from base classes
//////////

// public

void vec::print_elements(const print_context & c, const char *row_start, const char *row_end, const char *row_sep) const
{
	c.s << row_start;
	for (unsigned ro=0; ro<row; ++ro) {
	  m[ro].print(c);
	}
	c.s << row_end;
}

void vec::do_print(const print_context & c, unsigned level) const
{
	c.s << "[";
	print_elements(c, "[", "]", ",", ",");
	c.s << "]";
}

void vec::do_print_latex(const print_latex & c, unsigned level) const
{
	c.s << "\\left(\\begin{array}{" << std::string(size_,'c') << "}";
	print_elements(c, "", "", "\\\\", "&");
	c.s << "\\end{array}\\right)";
}

void vec::do_print_python_repr(const print_python_repr & c, unsigned level) const
{
	c.s << class_name() << '(';
	print_elements(c, "[", "]", ",", ",");
	c.s << ')';
}

/** nops is defined to be rows x columns. */
size_t vec::nops() const
{
	return static_cast<size_t>(size_); 
}

/** returns vec entry at position (i/col, i%col). */
ex vec::op(size_t i) const
{
	GINAC_ASSERT(i<nops());
	
	return m[i];
}

/** returns writable vec entry at position (i/col, i%col). */
ex & vec::let_op(size_t i)
{
	GINAC_ASSERT(i<nops());
	
	ensure_if_modifiable();
	return m[i];
}

/** Evaluate vec entry by entry. */
ex vec::eval(int level) const
{
	// check if we have to do anything at all
	if ((level==1)&&(flags & status_flags::evaluated))
		return *this;
	
	// emergency break
	if (level == -max_recursion_level)
		throw (std::runtime_error("vec::eval(): recursion limit exceeded"));
	
	// eval() entry by entry
	exvector m2(size_);
	--level;
	for (unsigned r=0; r<size_; ++r)
			m2[r] = m[r].eval(level);
	
	return (new vec(row, m2))->setflag(status_flags::dynallocated | status_flags::evaluated);
}

ex vec::subs(const exmap & mp, unsigned options) const
{
	exvector m2(row * col);
	for (unsigned r=0; r<row; ++r)
		for (unsigned c=0; c<col; ++c)
			m2[r*col+c] = m[r*col+c].subs(mp, options);

	return vec(row, col, m2).subs_one_level(mp, options);
}



// protected

int vec::compare_same_type(const basic & other) const
{
	GINAC_ASSERT(is_exactly_a<vec>(other));
	const vec &o = static_cast<const vec &>(other);
	
	// compare number of rows
	if (row != o.rows())
		return row < o.rows() ? -1 : 1;
	
	// compare number of columns
	if (col != o.cols())
		return col < o.cols() ? -1 : 1;
	
	// equal number of rows and columns, compare individual elements
	int cmpval;
	for (unsigned r=0; r<row; ++r) {
		for (unsigned c=0; c<col; ++c) {
			cmpval = ((*this)(r,c)).compare(o(r,c));
			if (cmpval!=0) return cmpval;
		}
	}
	// all elements are equal => matrices are equal;
	return 0;
}

bool vec::match_same_type(const basic & other) const
{
	GINAC_ASSERT(is_exactly_a<vec>(other));
	const vec & o = static_cast<const vec &>(other);
	
	// The number of rows and columns must be the same. This is necessary to
	// prevent a 2x3 vec from matching a 3x2 one.
	return size_== o.size();
}

/** Automatic symbolic evaluation of an indexed vec. */
ex vec::eval_indexed(const basic & i) const
{
	GINAC_ASSERT(is_a<indexed>(i));
	GINAC_ASSERT(is_a<vec>(i.op(0)));

	bool all_indices_unsigned = static_cast<const indexed &>(i).all_index_values_are(info_flags::nonnegint);

	// Check indices
	if (i.nops() == 2) {

		// One index, must be one-dimensional vector
		if (size_ != 1 )
			throw (std::runtime_error("vec::eval_indexed(): vector must have exactly 1 index"));

		const idx & i1 = ex_to<idx>(i.op(1));

		if (size_ == 1) {

			// Column vector
			if (!i1.get_dim().is_equal(row))
				throw (std::runtime_error("vec::eval_indexed(): dimension of index must match number of vector elements"));

			// Index numeric -> return vector element
			if (all_indices_unsigned) {
				unsigned n1 = ex_to<numeric>(i1.get_value()).to_int();
				if (n1 >= row)
					throw (std::runtime_error("vec::eval_indexed(): value of index exceeds number of vector elements"));
				return (*this)(n1, 0);
			}

		} 

	}  else
		throw (std::runtime_error("vec::eval_indexed(): vec must have exactly 2 indices"));

	return i.hold();
}

/** Sum of two indexed matrices. */
ex vec::add_indexed(const ex & self, const ex & other) const
{
	GINAC_ASSERT(is_a<indexed>(self));
	GINAC_ASSERT(is_a<vec>(self.op(0)));
	GINAC_ASSERT(is_a<indexed>(other));
	GINAC_ASSERT(self.nops() == 2 || self.nops() == 3);

	// Only add two matrices
	if (is_a<vec>(other.op(0))) {
		GINAC_ASSERT(other.nops() == 2 || other.nops() == 3);

		const vec &self_vec = ex_to<vec>(self.op(0));
		const vec &other_vec = ex_to<vec>(other.op(0));

		if (self.nops() == 2 && other.nops() == 2) { // vector + vector

			if (self_vec.row == other_vec.row)
				return indexed(self_vec.add(other_vec), self.op(1));
			else if (self_vec.row == other_vec.col)
				return indexed(self_vec.add(other_vec.transpose()), self.op(1));

		} else if (self.nops() == 3 && other.nops() == 3) { // vec + vec

			if (self.op(1).is_equal(other.op(1)) && self.op(2).is_equal(other.op(2)))
				return indexed(self_vec.add(other_vec), self.op(1), self.op(2));
			else if (self.op(1).is_equal(other.op(2)) && self.op(2).is_equal(other.op(1)))
				return indexed(self_vec.add(other_vec.transpose()), self.op(1), self.op(2));

		}
	}

	// Don't know what to do, return unevaluated sum
	return self + other;
}

/** Product of an indexed vec with a number. */
ex vec::scalar_mul_indexed(const ex & self, const numeric & other) const
{
	GINAC_ASSERT(is_a<indexed>(self));
	GINAC_ASSERT(is_a<vec>(self.op(0)));
	GINAC_ASSERT(self.nops() == 2 || self.nops() == 3);

	const vec &self_vec = ex_to<vec>(self.op(0));

	if (self.nops() == 2)
		return indexed(self_vec.mul(other), self.op(1));
	else // self.nops() == 3
		return indexed(self_vec.mul(other), self.op(1), self.op(2));
}

/** Contraction of an indexed vec with something else. */
bool vec::contract_with(exvector::iterator self, exvector::iterator other, exvector & v) const
{
	GINAC_ASSERT(is_a<indexed>(*self));
	GINAC_ASSERT(is_a<indexed>(*other));
	GINAC_ASSERT(self->nops() == 2 || self->nops() == 3);
	GINAC_ASSERT(is_a<vec>(self->op(0)));

	// Only contract with other matrices
	if (!is_a<vec>(other->op(0)))
		return false;

	GINAC_ASSERT(other->nops() == 2 || other->nops() == 3);

	const vec &self_vec = ex_to<vec>(self->op(0));
	const vec &other_vec = ex_to<vec>(other->op(0));

	if (self->nops() == 2) {

		if (other->nops() == 2) { // vector * vector (scalar product)

			if (self_vec.col == 1) {
				if (other_vec.col == 1) {
					// Column vector * column vector, transpose first vector
					*self = self_vec.transpose().mul(other_vec)(0, 0);
				} else {
					// Column vector * row vector, swap factors
					*self = other_vec.mul(self_vec)(0, 0);
				}
			} else {
				if (other_vec.col == 1) {
					// Row vector * column vector, perfect
					*self = self_vec.mul(other_vec)(0, 0);
				} else {
					// Row vector * row vector, transpose second vector
					*self = self_vec.mul(other_vec.transpose())(0, 0);
				}
			}
			*other = DUMMY;
			return true;

		} else { // vector * vec

			// B_i * A_ij = (B*A)_j (B is row vector)
			if (is_dummy_pair(self->op(1), other->op(1))) {
				if (self_vec.row == 1)
					*self = indexed(self_vec.mul(other_vec), other->op(2));
				else
					*self = indexed(self_vec.transpose().mul(other_vec), other->op(2));
				*other = DUMMY;
				return true;
			}

			// B_j * A_ij = (A*B)_i (B is column vector)
			if (is_dummy_pair(self->op(1), other->op(2))) {
				if (self_vec.col == 1)
					*self = indexed(other_vec.mul(self_vec), other->op(1));
				else
					*self = indexed(other_vec.mul(self_vec.transpose()), other->op(1));
				*other = DUMMY;
				return true;
			}
		}

	} else if (other->nops() == 3) { // vec * vec

		// A_ij * B_jk = (A*B)_ik
		if (is_dummy_pair(self->op(2), other->op(1))) {
			*self = indexed(self_vec.mul(other_vec), self->op(1), other->op(2));
			*other = DUMMY;
			return true;
		}

		// A_ij * B_kj = (A*Btrans)_ik
		if (is_dummy_pair(self->op(2), other->op(2))) {
			*self = indexed(self_vec.mul(other_vec.transpose()), self->op(1), other->op(1));
			*other = DUMMY;
			return true;
		}

		// A_ji * B_jk = (Atrans*B)_ik
		if (is_dummy_pair(self->op(1), other->op(1))) {
			*self = indexed(self_vec.transpose().mul(other_vec), self->op(2), other->op(2));
			*other = DUMMY;
			return true;
		}

		// A_ji * B_kj = (B*A)_ki
		if (is_dummy_pair(self->op(1), other->op(2))) {
			*self = indexed(other_vec.mul(self_vec), other->op(1), self->op(2));
			*other = DUMMY;
			return true;
		}
	}

	return false;
}


//////////
// non-virtual functions in this class
//////////

// public

/** Sum of matrices.
 *
 *  @exception logic_error (incompatible matrices) */
vec vec::add(const vec & other) const
{
	if (col != other.col || row != other.row)
		throw std::logic_error("vec::add(): incompatible matrices");
	
	exvector sum(this->m);
	exvector::iterator i = sum.begin(), end = sum.end();
	exvector::const_iterator ci = other.m.begin();
	while (i != end)
		*i++ += *ci++;
	
	return vec(row,col,sum);
}


/** Difference of matrices.
 *
 *  @exception logic_error (incompatible matrices) */
vec vec::sub(const vec & other) const
{
	if (col != other.col || row != other.row)
		throw std::logic_error("vec::sub(): incompatible matrices");
	
	exvector dif(this->m);
	exvector::iterator i = dif.begin(), end = dif.end();
	exvector::const_iterator ci = other.m.begin();
	while (i != end)
		*i++ -= *ci++;
	
	return vec(row,col,dif);
}


/** Product of matrices.
 *
 *  @exception logic_error (incompatible matrices) */
vec vec::mul(const vec & other) const
{
	if (this->cols() != other.rows())
		throw std::logic_error("vec::mul(): incompatible matrices");
	
	exvector prod(this->rows()*other.cols());
	
	for (unsigned r1=0; r1<this->rows(); ++r1) {
		for (unsigned c=0; c<this->cols(); ++c) {
			if (m[r1*col+c].is_zero())
				continue;
			for (unsigned r2=0; r2<other.cols(); ++r2)
				prod[r1*other.col+r2] += (m[r1*col+c] * other.m[c*other.col+r2]).expand();
		}
	}
	return vec(row, other.col, prod);
}


/** Product of vec and scalar. */
vec vec::mul(const numeric & other) const
{
	exvector prod(size_);

	for (unsigned r=0; r<size_; ++r)
	   prod[r] = m[r] * other;

	return vec(row, prod);
}


/** Product of vec and scalar expression. */
vec vec::mul_scalar(const ex & other) const
{
	if (other.return_type() != return_types::commutative)
		throw std::runtime_error("vec::mul_scalar(): non-commutative scalar");

	exvector prod(size_);

	for (unsigned r=0; r<row; ++r)
	  prod[r] = m[r] * other;

	return vec(row, prod);
}





/** operator() to access elements for reading.
 *
 *  @param ro row of element
 *  @param co column of element
 *  @exception range_error (index out of range) */
const ex & vec::operator() (unsigned ro, unsigned co) const
{
	if (ro>=row || co>=col)
		throw (std::range_error("vec::operator(): index out of range"));

	return m[ro*col+co];
}


/** operator() to access elements for writing.
 *
 *  @param ro row of element
 *  @param co column of element
 *  @exception range_error (index out of range) */
ex & vec::operator() (unsigned ro, unsigned co)
{
	if (ro>=row || co>=col)
		throw (std::range_error("vec::operator(): index out of range"));

	ensure_if_modifiable();
	return m[ro*col+co];
}


/** Transposed of an m x n vec, producing a new n x m vec object that
 *  represents the transposed. */
vec vec::transpose() const
{
	exvector trans(this->cols()*this->rows());
	
	for (unsigned r=0; r<this->cols(); ++r)
		for (unsigned c=0; c<this->rows(); ++c)
			trans[r*this->rows()+c] = m[c*this->cols()+r];
	
	return vec(this->cols(),this->rows(),trans);
}










ex lst_to_vec(const lst & l)
{
	lst::const_iterator itr, itc;

	// Find number of rows and columns
	size_t rows = l.nops(), cols = 0;
	for (itr = l.begin(); itr != l.end(); ++itr) {
		if (!is_a<lst>(*itr))
			throw (std::invalid_argument("lst_to_vec: argument must be a list of lists"));
		if (itr->nops() > cols)
			cols = itr->nops();
	}

	// Allocate and fill vec
	vec &M = *new vec(rows, cols);
	M.setflag(status_flags::dynallocated);

	unsigned i;
	for (itr = l.begin(), i = 0; itr != l.end(); ++itr, ++i) {
		unsigned j;
		for (itc = ex_to<lst>(*itr).begin(), j = 0; itc != ex_to<lst>(*itr).end(); ++itc, ++j)
			M(i, j) = *itc;
	}

	return M;
}



} // namespace GiNaC
