import os.path
from SCons.Options import Options

class CommandError(Exception):
    def __init__(self, command, exitVal, stderr):
        Exception.__init__(self, "Command `%s' exited with value: %d" % (command, exitVal))
        self.stderr = stderr

def runCommand(command, args):
    """ Run command, returning its output on the stdout and stderr streams.

    If the command exits with a non-zero value, CommandError is raised.
    @param command: The name of the command
    @param args: The arguments to the command, either as an explicit sequence or as a whitespace delimited string.
    @return: A pair of strings, containing the command's output on stdout and stderr, respectively.
    """
    if isinstance(args, basestring):
        args = args.split()
    from popen2 import Popen3
    cl = "%s %s" % (command, " ".join(args))
    p = Popen3(cl, capturestderr=True)
    r = p.wait()
    try:
        out = p.fromchild.read().strip()
        err = p.childerr.read().strip()
        if r:
            raise CommandError(cl, r, err)
    finally:
        p.tochild.close()
        p.fromchild.close()
        p.childerr.close()

    return out, err

def rsplit(toSplit, sub, max):
    """ str.rsplit seems to have been introduced in 2.4 :( """
    l = []
    i = 0
    while i < max:
        idx = toSplit.rfind(sub)
        if idx != -1:
            toSplit, splitOff = toSplit[:idx], toSplit[idx + len(sub):]
            l.insert(0, splitOff)
            i = i + 1
        else:
            break

    l.insert(0, toSplit)
    return l

class Configure(object):
    """ Replace the standard SCons Configure method.
    
    The standard SCons Configure method is somewhat weak, in particular with regard to cleaning up after
    previous builds. When cleaning you normally don't want to reconfigure, you just want to clean the
    targets from the previous build. You may still want to get at configuration parameters however,
    since they may affect which targets are built (e.g. a module is only built in the presence of an
    external library). This class caches configuration values between builds, even command-line options,
    so that the build configuration when cleaning mirrors that when building.
    """
    def __init__(self, env, arguments, options=[]):
        """ @param env: SCons environment
        @param arguments: SCons command-line arguments (ARGUMENTS)
        @param options: Command-line options.
        """
        useCache = env.GetOption("clean") # Prefer cached values if cleaning
        
        self._conf = env.Configure(log_file=os.path.join("scons", "configure.log"))
        self._oldCache, self._newCache = {}, {}
        self._cachePath = os.path.abspath(os.path.join("scons", "configure.cache"))
        if useCache and os.path.exists(self._cachePath):
            try:
                f = open(self._cachePath)
                context = None
                import re
                for l in f:
                    m = re.match(r"(.+):", l)
                    if m:
                        context = m.group(1)
                        self._oldCache[context] = {}
                        continue
                    if context is None:
                        continue

                    m = re.match("(.+) = (.+)", l)
                    try:
                        k, v = m.groups()
                    except:
                        continue
                    self._oldCache[context][k] = v
            finally:
                f.close()

        args = arguments.copy()
        optsCache = os.path.join("scons", "options.cache")   # Save options between runs in this cache
        if useCache:
            # Then ignore new values
            for o in options:
                if o[0] in args:
                    del args[o[0]]

        opts = Options(optsCache, args=args)
        opts.AddOptions(*options)
        opts.Update(env)
        # Cache options if asked to
        cacheOptions = env.get("cacheOptions", False)
        if cacheOptions:
            del env["cacheOptions"] # Don't store this value
            opts.Save(optsCache, env)
            # Restore
            env["cacheOptions"] = cacheOptions
        env.Help(opts.GenerateHelpText(env))
    
    def checkCxxHeader(self, header, include_quotes='""'):
        ctxt = "checkCxxHeader"
        try:
            r = self._oldCache[ctxt][str(header)]
        except KeyError:
            r = self._conf.CheckCXXHeader(header, include_quotes)

        if not ctxt in self._newCache:
            self._newCache[ctxt] = {}
        self._newCache[ctxt][str(header)] = r
        return r

    def checkLibWithHeader(self, lib, header, language="C", call=None):
        ctxt = "checkLibWithHeader"
        try:
            r = self._oldCache[ctxt][str(lib)]
        except KeyError:
            kwds = {}
            if call is not None:
                kwds["call"] = call
            r = self._conf.CheckLibWithHeader(lib, header, language, **kwds)

        if not ctxt in self._newCache:
            self._newCache[ctxt] = {}
        self._newCache[ctxt][str(lib)] = r
        return r

    def finish(self):
        """ Update configuration cache """
        f = open(self._cachePath, "w")
        try:
            for ctxt, d in self._newCache.items():
                f.write("%s:\n" % ctxt)
                for k, v in d.items():
                    f.write("%s = %s\n" % (k, v))
        finally:
            f.close()

class PkgconfigError(Exception):
    def __init__(self, pkg, msg):
        Exception.__init__(self, msg)
        import string
        self.lib = rsplit(pkg, "-", 1)[0]

class PkgconfigMissing(PkgconfigError):
    def __init__(self, pkg):
        PkgconfigError.__init__(self, pkg, "The library '%s' couldn't be found, consider creating a pkg-config entry for it (`%s')" %
                (pkg.split("-", 1)[0], pkg + ".pc"))

def _pkgconfig(lib, param):
    try: runCommand("pkg-config", ["--exists", lib])
    except CommandError:
        raise PkgconfigMissing(lib)
    try: out, err = runCommand("pkg-config", [param, lib])
    except CommandError, err:
        raise PkgconfigError(lib, "Error reported by pkg-config: `%s'" % err.stderr)
    
    return out

def pkgconfigIncludeDirs(lib):
    out = _pkgconfig(lib, "--cflags-only-I")
    dirs = []
    opts = out.split()
    for o in opts:
        dirs.append(o[2:])
    return dirs

def pkgconfigLibDirs(lib):
    out = _pkgconfig(lib, "--libs-only-L")
    dirs = []
    opts = out.split()
    for o in opts:
        dirs.append(o[2:])
    return dirs

def pkgconfigLibs(lib):
    out = _pkgconfig(lib, "--libs-only-l")
    dirs = []
    opts = out.split()
    for o in opts:
        dirs.append(o[2:])
    return dirs
